(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NFCL_OBJ.PAS, unit for the nefclass object and methods        *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 01/22/96                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

Unit NFCL_OBJ;

interface

uses Dos, Crt, Graph, Drivers, App, Objects, Dialogs, TextView,
     Views, MsgBox, Memory,
     NFCL_Con, NFCL_Typ, NFCL_Tsk;

type

  (*-----------------------------------------------------------------------*)
  (*         NEFCLASS: A fuzzy perceptron for crisp classification         *)
  (*                                                                       *)
  (* A NEFCLASS network consists of 3 layers: input, rule and ouput layer. *)
  (* The antecedents are index values pointing to a fuzzy set of the       *)
  (* antecedent database AntecBase that stores the partionings of the input*)
  (* variables. The conclusion weights are real values from [0,1] indicat- *)
  (* ing a degree of support for the respective rule. Each rule unit is    *)
  (* connected to exactly one output unit.                                 *)
  (* The rule layer is empty at the beginning. Rule units are created      *)
  (* during the first pass through the data.                               *)
  (*-----------------------------------------------------------------------*)
  PNefclass = ^TNefclass;
  TNefclass = object (Tobject)
     (* Number of units per in layers 1, 2, and 3 *)
     L1, L2, L3                   : integer;
     (* Max. number of fuzzy rules to be created *)
     max_rules                    : integer;
     (* Activations of the input and ouput units *)
     InLayer, OutLayer            : UnitVector;
     (* Activations of the rule units *)
     RuleLayer                    : RuleVector;
     (* Antcedents: indeces to fuzzy sets *)
     Antecedents                  : WeightMatrix;
     (* database of fuzzy sets *)
     AntecBase                    : Farray;
     (* degrees of support for the rules *)
     ConclusionWeights            : WeightVector;
     (* Names of the output units (classes) *)
     ClassNames                   : array[1..maxunits] of VNStr;
     (* Ranges for the input units *)
     InVarRange                   : array[1..maxunits] of Range;
     (* indeces of fuzzy sets with minimal membership *)
     (* degree per rule unit *)
     Min_Ant                      : ByteVector;
     UseSum                       : Boolean;
     Individual                   : Boolean;
     LearnRules                   : Boolean;
     Saved                        : Boolean;
     Epochs                       : LongInt;
     Name                         : FNameStr;
     FileName                     : NameStr;
     NetLabel                     : LabelStr;

     constructor Init (AL1, AMax, AL3, ANoFS: integer;
                       UList, CList: PCollection;
                       IndividualFlag, UseSumFlag: Boolean);
     procedure InitNet (ATask: PTask);
     procedure ChangeNet (AMax, ANoFS: integer;
                          UList, CList: PCollection;
                          IndividualFlag, UseSumFlag: Boolean);
     destructor Done; virtual;
     function ReadNet (var ParamFile: Text; AName: FNameStr): Boolean;
     procedure WriteNet (var ParamFile: Text; AName: FNameStr;
                         Comment: string);
     procedure WriteRuleBase (var ParamFile: Text);
     function Membership (FS: TriangleFSet; x: real): real;
     procedure Defuzzify (var Nom, Denom: real; Match: real;
                          FS: TriangleFSet);
     procedure Propagate (APattern: PPattern; var min_input: UnitVector);
     procedure Free_Check_and_Adapt
                (var FS: TriangleFSet;
                 x, da, db, dc, minrange, maxrange: real; asym: boolean);
     procedure Constrained_Check_and_Adapt (R: integer; da, db, dc: real;
                                       var min_input, max_input: UnitVector);
     procedure NotPass_Check_and_Adapt
                               (R: integer; da, db, dc: real;
                                var min_input, max_input: UnitVector;
                                asym: boolean);
     procedure CreateRuleBase (var created, wanted: integer; T: PTask;
                               restricted: boolean);
     function  CheckOutput (APattern: PPattern; var Class: Integer;
                            var Hit: real; var ambiguous: boolean): Boolean;
     procedure ReconnectRules (T: PTask; var EFile: Text);
     procedure EvaluateRules (T: PTask; var EFile: Text;
                              RLType: RuleLearningType);
     function  FEBP (APattern: PPattern;
                     var min_input, max_input: UnitVector;
                     LP: PRLearnParam): real;
     procedure Testing (Task: PTask);
     procedure DisplayFuzzyWeights;
  end; (* TNefclass *)

implementation

(*=========================== Methods of TNefclass ========================*)

constructor TNefclass.Init (AL1, AMax, AL3, ANoFS: integer;
                            UList, CList: PCollection;
                            IndividualFlag, UseSumFlag: Boolean);
 var
  i : integer;
 begin
  L1   := AL1;
  L2   := 0;
  L3   := AL3;
  max_rules := AMax;
  Individual := IndividualFlag;
  for i := 1 to L1 do
   begin
    if UList <> nil
      then AntecBase[i].VName := PPartition(UList^.At(i-1))^.VName;
    if Individual and (UList <> nil)
      then AntecBase[i].count := PPartition(UList^.At(i-1))^.count
      else AntecBase[i].count := ANoFS;
    InVarRange[i].min := 1;
    InVarRange[i].max := -1;
   end;
  for i := 1 to L3 do
   if CList <> nil
    then ClassNames[i] := PVNStr(CList^.At(i-1))^;
  UseSum := UseSumFlag;
  Epochs := 0;
  LearnRules := False;
  Saved := False;
  Name := '';
  NetLabel := '';
  FileName := '';
  InitNet (nil);
 end; (* TNefclass.Init *)

(*-------------------------------------------------------------------------*)
(* Initialize a Nefclass network: Create a uniform partioning for each     *)
(* variable, set activations and conclusion weights to 0                   *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.InitNet (ATask: PTask);
 var
  i,j: integer;
  x  : real;

 begin

  if ATask <> nil
    then begin
     LearnRules := True;
     for i := 1 to L1 do begin
      InVarRange[i].min := ATask^.min_input[i];
      InVarRange[i].max := ATask^.max_input[i];
     end;
    end;

  (* create a uniform partitioning on the domain of each input variable *)
  for i := 1 to L1 do begin
   x := (InVarRange[i].max - InVarRange[i].min) / (AntecBase[i].count + 1);
   (* create special fuzzy set, in case a variable is not used in a rule *)
   AntecBase[i].FSets[0].a := InVarRange[i].min;
   AntecBase[i].FSets[0].b := InVarRange[i].min;
   AntecBase[i].FSets[0].c := InVarRange[i].max;
   AntecBase[i].FSets[0].right_shouldered := true;
   AntecBase[i].FSets[0].left_shouldered := true;
   GetFName (0,1, AntecBase[i].FSets[0].Name);

   for j := 1 to AntecBase[i].count  do begin
    if j > 1
     then begin
      AntecBase[i].FSets[j].a := AntecBase[i].FSets[j-1].b;
      AntecBase[i].FSets[j].left_shouldered := false;
      AntecBase[i].FSets[j].right_shouldered := false;
     end
     else begin
      AntecBase[i].FSets[j].a := InVarRange[i].min;
      AntecBase[i].FSets[j].left_shouldered := true;
      AntecBase[i].FSets[j].right_shouldered := false;
     end;
    AntecBase[i].FSets[j].b := AntecBase[i].FSets[j].a + x;
    AntecBase[i].FSets[j].c := AntecBase[i].FSets[j].b + x;
    GetFName (AntecBase[i].count,j, AntecBase[i].FSets[j].Name);
   end; (* for j *)
   AntecBase[i].FSets[AntecBase[i].count].right_shouldered := true;
  end; (* for i *)

  for i := 1 to maxunits do
   begin
    InLayer[i] := 0.0; RuleLayer[i] := 0.0; OutLayer[i] := 0.0;
    Min_Ant[i] := 0;
    if i > L2
     then begin
      ConclusionWeights[i].weight := 0.0;
      ConclusionWeights[i].node := 0;
     end;
   end; (* for i *)

 end; (* TNefclass.InitNet *)

(*-------------------------------------------------------------------------*)
(* Change a Nefclass network: Create a new uniform partioning for each     *)
(* variable, whose number of fuzzy sets was changed.                       *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.ChangeNet (AMax, ANoFS: integer;
                               UList, CList: PCollection;
                               IndividualFlag, UseSumFlag: Boolean);
 var
  i,j: integer;
  x  : real;
  HasChanged: Boolean;
  NewNoFS: integer;
  RuleFlag: Boolean;

 begin

  Saved := False;

  if max_rules <> AMax
   then begin
    LearnRules := True;
    max_rules  := AMax;
   end;

  Individual := IndividualFlag;
  UseSum     := UseSumFlag;
  RuleFlag   := False;

  for i := 1 to L3 do
   if CList <> nil
    then ClassNames[i] := PVNStr(CList^.At(i-1))^;

  for i := 1 to L1 do begin
   if UList <> nil
     then AntecBase[i].VName := PPartition(UList^.At(i-1))^.VName;
   if Individual and (UList <> nil)
     then NewNoFS := PPartition(UList^.At(i-1))^.count
     else NewNoFS := ANoFS;

   if AntecBase[i].count <> NewNoFS
     then begin
      AntecBase[i].count := NewNoFS;
      HasChanged := True;
      RuleFlag := L2 > 0;
      LearnRules := L2 > 0;
      L2 := 0; (* The rules must be deleted, if a partitioning changes !*)
     end
     else HasChanged := False;

   if HasChanged
     then begin (* create new partitioning *)
      x := (InVarRange[i].max - InVarRange[i].min) /
           (AntecBase[i].count + 1);
      for j := 1 to AntecBase[i].count  do begin
       if j > 1
        then begin
         AntecBase[i].FSets[j].a := AntecBase[i].FSets[j-1].b;
         AntecBase[i].FSets[j].left_shouldered := false;
         AntecBase[i].FSets[j].right_shouldered := false;
        end
        else begin
         AntecBase[i].FSets[j].a := InVarRange[i].min;
         AntecBase[i].FSets[j].left_shouldered := true;
         AntecBase[i].FSets[j].right_shouldered := false;
        end;
       AntecBase[i].FSets[j].b := AntecBase[i].FSets[j].a + x;
       AntecBase[i].FSets[j].c := AntecBase[i].FSets[j].b + x;
       GetFName (AntecBase[i].count,j, AntecBase[i].FSets[j].Name);
      end; (* for j *)
      AntecBase[i].FSets[AntecBase[i].count].right_shouldered := true;
     end; (* If HasChanged then *)
  end; (* for i *)
  if RuleFlag
   then MessageBox ('For at least one variable the number of fuzzy sets '+
                    'was changed. Therefore all rules were deleted.', nil,
                    mfInformation or mfOKButton);

 end; (* TNefclass.ChangeNet *)

(*----------------------- Destructor of TNefclass -------------------------*)

destructor TNefclass.Done;
 begin
  inherited Done;
 end; (* TNefclass.Done *)

(*-------------------------------------------------------------------------*)
(* Read a previously defined fuzzy perceptron from a textfile and return   *)
(* TRUE if network was read successfully.                                  *)
(* Format: [<comment>] <parameters> [[<comment>] <data area>] ... END      *)
(*   <comment>   :  lines beginning with %                                 *)
(*   <parameters>:  PARAMETERS <newline>                                   *)
(*                  <L1> <L2> <L3> <max_rules> <AntecBase.count array>     *)
(*                  <Ranges of input variables: array of <min max> pairs>  *)
(*                  <newline>                                              *)
(*                  [USESUM <newline> | USEMAXIMUM <newline>]              *)
(*                  [INDIVIDUAL <newline> | SAME <newline>]                *)
(*                  [VNAMES <newline> <var_names>]                         *)
(*   <var_names>:   <<string> <newline> <string> <newline> ...>            *)
(*   <data area>:   <fuzzy sets> | <connections> | <weights>               *)
(*   <fuzzy sets>:  [NAMES <newline>] FUZZY <newline>                      *)
(*                  <<a b c ls rs [n]> <a b c ls rs [n]> ...> <newline>    *)
(*   <connections>: MATRIX <newline>                                       *)
(*                  <Antecedents array> <newline>                          *)
(*   <weights>:     WEIGHTS <newline>                                      *)
(*                  <ConclusionWeights array> <newline>                    *)
(*   <a b c ls rs [n]>: a = left point, b = peak, c = right point,         *)
(*                      ls = 1 if left shouldered, rs = 1 if right         *)
(*                      shouldered), n is an optional name (will be only   *)
(*                      read if command NAMES is specified before FUZZY)   *)
(*-------------------------------------------------------------------------*)

function TNefclass.ReadNet (var ParamFile: Text; AName: FNameStr): Boolean;

 var
  command     : string;
  Result      : boolean;
  TheEnd      : boolean;
  buildnames  : boolean;
  dovnames    : boolean;
  Dir         : DirStr;
  Ext         : ExtStr;

  (*--- internal procedure to read the net structure parameters ---*)
  procedure read_param;
   var
    i : integer;
   begin
     read (Paramfile, L1, L2, L3, max_rules);
     for i := 1 to L1 do
      read (Paramfile, AntecBase[i].count);
     readln (Paramfile);
     for i := 1 to L1 do
      read (Paramfile, InVarRange[i].min, InVarRange[i].max);
     readln (Paramfile);
   end; (* read_param *)

  (*- internal procedure to read the names of input and output variables -*)
  procedure read_vnames;
   var
    i : integer;
   begin
    for i := 1 to L1 do
     readln (Paramfile, Antecbase[i].VName);
    for i := 1 to L3 do
     readln (Paramfile, ClassNames[i]);
    dovnames := false;
   end;

  (*- internal procedure to create names of input and output variables -*)
  procedure create_vnames;
   var
    i : integer;
    S : VNStr;
   begin
    for i := 1 to L1 do
     begin
      str (i,S);
      Antecbase[i].VName := 'V_' + S;
     end;
    for i := 1 to L3 do
     begin
      str (i,S);
      ClassNames[i] := 'Class_'+ S;
     end;
   end;

  (*--- internal procedure to read the fuzzy weights of the antecedents ---*)
  procedure read_fuzzy;
   var
    i, j, ls, rs : integer;
    blank        : char;
   begin
     for i := 1 to L1 do
      with AntecBase[i] do
       begin
        (* create special fuzzy set, in case a variable is not used in a rule *)
        FSets[0].a := InVarRange[i].min;
        FSets[0].b := InVarRange[i].min;
        FSets[0].c := InVarRange[i].max;
        FSets[0].right_shouldered := true;
        FSets[0].left_shouldered := true;
        GetFName (0,1, FSets[0].Name);

        for j := 1 to count do
         begin
          if buildnames
            then begin
             readln (Paramfile, FSets[j].a, FSets[j].b, FSets[j].c, ls, rs);
             GetFName (count, j, FSets[j].Name);
            end
            else begin
             read (Paramfile, FSets[j].a, FSets[j].b, FSets[j].c, ls, rs);
             readln (Paramfile, blank, FSets[j].Name);
            end;
          FSets[j].left_shouldered  := (ls = 1);
          FSets[j].right_shouldered := (rs = 1);
         end; (* for j *)
       end; (* with *)
    readln (ParamFile);
   end; (* read_fuzzy *)

   (*--- internal procedure to read the connection matrix ---*)
   procedure read_matrix;
    var
     i,j : integer;
    begin
      for i := 1 to L1 do
       for j := 1 to L2 do
        read (ParamFile, Antecedents[i,j]);
      readln (ParamFile);
    end; (* read_matrix *)

   (*--- internal procedure to read the conclusion weights ---*)
   procedure read_weights;
    var
     j : integer;
    begin
      for j := 1 to L2 do
       read (ParamFile, ConclusionWeights[j].node,
                        ConclusionWeights[j].weight);
      readln (ParamFile);
    end; (* read_weights *)

   (*--- internal function to give error message and to return FALSE ---*)
   function error: boolean;
    begin
      MessageBox ('Unexpected error in function ReadNet', nil,
                   mfError or mfOKButton);
     error := FALSE;
    end;

 (*------- Body of ReadNet ---------*)
 begin

  Reset (ParamFile);
  Result := True;
  TheEnd := False;
  buildnames := true;
  dovnames := true;
  UseSum   := true;
  Individual := true;
  LearnRules := False;
  Saved      := True;
  Name       := AName;
  FSplit(AName, Dir, FileName, Ext);
  NetLabel   := '';

  repeat
   command := skip_comments (ParamFile);
   if pos ('P', command) = 1
    then read_param
   else if pos ('USES', command) = 1
     then UseSum := True
   else if pos ('USEM', command) = 1
     then UseSum := False
   else if pos ('I', command) = 1
     then Individual := True
   else if pos ('S', command) = 1
     then Individual := False
   else if pos ('V', command) = 1
     then read_vnames
   else if pos ('N', command) = 1
     then buildnames := false
   else if pos ('F', command) = 1
     then read_fuzzy
   else if pos ('M', command) = 1
     then read_matrix
   else if pos ('W', command) = 1
     then read_weights
   else if pos ('END', command) = 1
     then TheEnd := True
   else Result := error;
  until TheEnd or (Result = FALSE);

  Close (ParamFile);
  if dovnames then create_vnames;
  ReadNet := Result;

 end; (* TNefclass.ReadNet *)

(*-------------------------------------------------------------------------*)
(* Write a Fuzzy Perceptron to a textfile (format see TNefclass.ReadNet).  *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.WriteNet (var ParamFile: Text; AName: FNameStr;
                              Comment: string);
 var
  i, j   : integer;
  ls, rs : integer;
  h,m,s,ss: word; (* to get the time *)
  min     : string[2];
  y,mo,d,w : word; (* to get the date *)
  Dir         : DirStr;
  Ext         : ExtStr;

 begin

  Name := AName;
  FSplit(AName, Dir, FileName, Ext);
  Rewrite (ParamFile);

  GetDate (y,mo,d,w); GetTime (h,m,s,ss);
  str (m:2, min);
  if m < 10 then min[1] := '0';

  writeln (ParamFile, '% NEFCLASS network file created by ',
                      'NEFCLASS-PC ', VersionString,
                      ', (c) TU Braunschweig, 1994, 1995');
  writeln (ParamFile, '% Filename: ', Name);
  writeln (ParamFile, '% This file was created on ', mo,'/',d,'/',y, ' at ',
                      h,':',min, '.');
  writeln (ParamFile, '%');
  writeln (ParamFile, '% ', Comment);
  writeln (ParamFile);


   (* Parameters: writing number of units and max. number of rules *)
   writeln (ParamFile, '% This are the network parameters');
   writeln (ParamFile, 'PARAMETERS');
   writeln (ParamFile, L1:10, L2:10, L3:10, max_rules:10);
   for i := 1 to L1 do begin
    write (ParamFile, AntecBase[i].count:10);
    if ((i mod 7) = 0) or (i = L1) then writeln (ParamFile);
   end; (* for i *)
   for i := 1 to L1 do begin
    write (ParamFile, InVarRange[i].min:10:4, ' ',
                      InVarRange[i].max:10:4, ' ');
    if ((i mod 3) = 0) or (i = L1) then writeln (ParamFile);
   end; (* for i *)

   writeln (ParamFile);

   if UseSum
    then writeln (ParamFile, 'USESUM')
    else writeln (ParamFile, 'USEMAXIMUM');
   writeln (ParamFile);

   if Individual
    then writeln (ParamFile, 'INDIVIDUAL')
    else writeln (ParamFile, 'SAME');
   writeln (ParamFile);

   writeln (ParamFile, '% This are the names of the input and output units');
   writeln (ParamFile, 'VNAMES');
   for i := 1 to L1 do
    writeln (ParamFile, Antecbase[i].VName);
   for i := 1 to L3 do
    writeln (ParamFile, ClassNames[i]);
   writeln (ParamFile);

   (* Fuzzy: writing parameters of the fuzzy sets *)
   writeln (ParamFile, '% This are the parameters of the fuzzy sets');
   writeln (ParamFile, 'NAMES');
   writeln (ParamFile, 'FUZZY');
   for i := 1 to L1 do
    with AntecBase[i] do
     for j := 1 to count do (* antecedents *)
       with FSets[j] do begin
        if left_shouldered
          then ls := 1
          else ls := 0;
        if right_shouldered
          then rs := 1
          else rs := 0;
        writeln (ParamFile, a:14:7, ' ', b:14:7, ' ', c:14:7,
                            ls: 3, rs: 3, ' ', Name);
       end; (* with FSets[j] *)
   writeln (ParamFile);

   (* Matrix: writing the connection matrix *)
   writeln (ParamFile, '% This is the connection matrix');
   writeln (ParamFile, 'MATRIX');
   for i := 1 to L1 do begin
    for j := 1 to L2 do
     write (ParamFile, Antecedents[i,j]: 4);
    writeln (ParamFile);
   end; (* for i *)
   writeln (ParamFile);

   (* Weights: writing the conclusion weights *)
   writeln (ParamFile, '% This are the conclusion weights');
   writeln (ParamFile, 'WEIGHTS');
   for i := 1 to L2 do begin
    write (ParamFile, ConclusionWeights[i].node:4,
                      ConclusionWeights[i].weight:14:7, ' ');
    if (i mod 4 = 0) or (i = L2) then writeln (ParamFile);
   end; (* for i *)
   writeln (ParamFile);

   writeln (ParamFile, 'END');

  Close (ParamFile);
  Saved := True;

 end; (* TNefclass.WriteNet *)

(*-------------------------------------------------------------------------*)
(* Procedure to write the rule base of the network into a text file        *)
(*-------------------------------------------------------------------------*)

procedure TNefclass.WriteRuleBase (var ParamFile: Text);
 var
  i, j   : integer;
  ls, rs : integer;
  h,m,s,ss: word; (* to get the time *)
  min     : string[2];
  y,mo,d,w : word; (* to get the date *)

 begin

  GetDate (y,mo,d,w); GetTime (h,m,s,ss);
  str (m:2, min);
  if m < 10 then min[1] := '0';

  writeln (ParamFile, '% Rule base file created by ',
                      'NEFCLASS ', VersionString,
                      ', (c) TU Braunschweig, 1994');
  writeln (ParamFile, '% This file was created on ', mo,'/',d,'/',y, ' at ',
                      h,':',min, '.');
  writeln (ParamFile, '%');
  writeln (ParamFile);
  for i := 1 to L2 do
   begin
    write (ParamFile, 'IF ');
    for j := 1 to L1 do
     begin
      write (ParamFile, AntecBase[j].VName:8, ' is ');
      write (ParamFile, AntecBase[j].FSets[Antecedents[j,i]].Name:8);
      if j < L1
        then begin
         write (ParamFile, ' and ');
         if (j mod 3 = 0)
           then begin
            writeln (ParamFile);
            write (ParamFile, '   ');
           end;
        end;
     end;
    writeln (ParamFile);
    writeln (ParamFile, 'THEN ', ClassNames[ConclusionWeights[i].Node]:8);
    writeln (ParamFile);
   end; (* for *)

 end; (* TNefclass.WriteRuleBase *)

(*-------------------------------------------------------------------------*)
(* Calculate the degree of membership of value x and fuzzy set FS. FS is   *)
(* a triangular function that might be shouldered.                         *)
(*-------------------------------------------------------------------------*)
function TNefclass.Membership (FS: TriangleFSet; x: real): real;
 begin
  if (x = FS.b)
    then Membership := 1.0
    else
     if (x >= FS.a) and (x < FS.b)
       then if FS.left_shouldered
               then Membership := 1.0
               else Membership := (x - FS.a) / (FS.b - FS.a)
       else
        if (x > FS.b) and (x <= FS.c)
          then if FS.right_shouldered
                 then Membership := 1.0
                 else Membership := (FS.c - x) / (FS.c - FS.b)
          else if ( (x < FS.a) and FS.left_shouldered ) or
                  ( (x > FS.c) and FS.right_shouldered)
                 then Membership := 1.0
                 else Membership := 0.0;
 end; (* TNefclass.Membership *)

(*-------------------------------------------------------------------------*)

procedure TNefclass.Defuzzify (var Nom, Denom: real; Match: real;
                               FS: TriangleFSet);
 var
  T1, T2: real;
  D, N: real;
 begin
  if (FS.c - FS.a) < mindiff
    then begin
     Nom := Nom + FS.b * Match;
     Denom := Denom + Match;
    end
    else begin
     T1 := Match * (FS.b - FS.a) + FS.a;
     T2 := FS.c - Match * (FS.c - FS.b);
     D :=  Match / 2 * (FS.c - FS.a + T2 - T1);
     N :=  1 / (FS.b - FS.a) * (  1/3 * T1 * T1 * T1
                                - 0.5 * FS.a * T1 * T1
                                + 1/6 * FS.a * FS.a * FS.a)
           + 0.5 * Match * (T2 * T2 - T1 * T1) +
           + 1 / (FS.c - FS.b) * (  1/6 * FS.c * FS.c * FS.c
                                  - 0.5 * FS.c * T2 * T2
                                  + 1/3 * T2 * T2 * T2);
     if (D < mindiff) or  (N < mindiff)
       then begin
        D := Match; N := Match * FS.b;
       end;
     Denom := Denom + D;
     Nom   := Nom   + N;
    end;
 end; (* TNefclass.Defuzzify *)

(*-------------------------------------------------------------------------*)
(* This procedure propagates a pattern through the fuzzy perceptron.       *)
(* The activation of a rule unit is determined by the minimum over the     *)
(* membership values coming from the antecedents, and the output values are*)
(* calculated by weighted sum of the rule activations.                     *)
(* If a missing value is encountered, then a membership value of 1.0 is    *)
(* assumed.                                                                *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.Propagate (APattern: PPattern;
                               var min_input: UnitVector);

 var
  i, j        : integer;
  Minimum     : real;
  MBV         : real;
  Nominator   : real;
  Denominator : real;
  Maximum     : real;

 begin

    for i := 1 to L1 do
      InLayer[i] := APattern^.InPattern[i];

    for j := 1 to L2 do
     begin
      Minimum := 2;
      Min_Ant[j] := 0;
      i := 1;
      while  (i <= L1) and (Minimum > 0) do
       begin
        if (InLayer[i] < min_input[i])
           then begin
            MBV := 1.0; (* for missing values, does not work yet *)
           end
           else MBV := Membership (AntecBase[i].FSets[Antecedents[i,j]],
                                   InLayer[i]);
        if MBV < Minimum
         then  begin
          Minimum := MBV;
          Min_Ant[j] := i;
         end; (* then *)
        inc (i);
       end; (* while *)
      RuleLayer[j] := Minimum;
     end; (* for j *)

    if UseSum
      then begin
       for j := 1 to L3 do begin
         Nominator := 0;
         Denominator := 0;
         for i := 1 to L2 do begin
          if ConclusionWeights[i].node = j
            then begin
              Nominator := Nominator + ConclusionWeights[i].weight
                           * RuleLayer[i];
              Denominator := Denominator + ConclusionWeights[i].weight;
              if RuleLayer[i] > Maximum
                then Maximum := RuleLayer[i];
            end; (* then *)
         end; (* for i *)
        if Denominator = 0
          then OutLayer[j] := 0
          else OutLayer[j] := Nominator / Denominator;
       end (* for j *)
      end (* if UseSum then *)
      else begin
       for j := 1 to L3 do begin
         Maximum := -1;
         for i := 1 to L2 do begin
          if ConclusionWeights[i].node = j
            then begin
              if RuleLayer[i] > Maximum
                then Maximum := RuleLayer[i];
            end; (* then *)
         end; (* for i *)
         OutLayer[j] := Maximum;
       end; (* for j *)
      end (* if UseSum ... else *)

 end; (* TNefclass.Propagate *)

(*-------------------------------------------------------------------------*)
(* This procedure changes the parameters of a fuzzy set, if the changes do *)
(* not conflict with the constraints: a < b < c, and b - a > minwidth, and *)
(* c - b > minwidth. The parameters must also stay within the domain.      *)
(* Changing one fuzzy set does not influence the other fuzzy sets.         *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.Free_Check_and_Adapt
            (var FS: TriangleFSet; x, da, db, dc, minrange, maxrange: real;
             asym: boolean);
 begin
  with FS do begin
   if ((a + da) < minrange) or ((a + da) > b - minwidth) or left_shouldered
      or (asym and (x >= b))
     then da := 0;
   if ((c + dc) > maxrange) or ((c + dc) < b + minwidth) or right_shouldered
      or (asym and (x <= b))
     then dc := 0;
   if ((a + da + minwidth) > (b + db)) or ((c + dc - minwidth) < (b + db))
     then db := 0;
   a := a + da;
   b := b + db;
   c := c + dc;
  end; (* with *)
 end; (* TNefclass.Free_Check_and_Adapt *)

(*-------------------------------------------------------------------------*)
(* This procedure makes sure that after changing a fuzzy set the neigh-    *)
(* boring fuzzy sets are also changed to insure that the membership values *)
(* add up to 1 at all points.                                              *)
(* The parameters of a fuzzy set are only changed, if the changes do       *)
(* not conflict with the constraints: a < b < c, and b - a > minwidth, and *)
(* c - b > minwidth. The parameters must also stay within the domain, and  *)
(* c (a) may not pass c (a) of its right (left) neighbor.                  *)
(* Only that half of the fuzzy set is changed, where the input value lies, *)
(* i.e. either a or c is changed, but not both.                            *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.Constrained_Check_and_Adapt
                               (R: integer; da, db, dc: real;
                                var min_input, max_input: UnitVector);
 var
  FS_l, FS, FS_r : ^TriangleFSet;
  V : integer;
  X : real;

 begin

  V := Min_Ant[R];

  (* Select fuzzy set FS and its left and right neighbors, if they exist *)
  if Antecedents[V,R] > 1
    then FS_l := @(AntecBase[V].FSets[Antecedents[V,R]-1])
    else FS_l := Nil;
  FS   := @(AntecBase[V].FSets[Antecedents[V,R]]);
  if Antecedents[V,R] < AntecBase[V].count
    then FS_r := @(AntecBase[V].FSets[Antecedents[V,R]+1])
    else FS_r := Nil;

  X := InLayer[V];

  (* we have no problems with neighbors here, because the last FS is *)
  (* right_shouldered, and the first FS is left_shouldered           *)
  with FS^ do begin
   if (X > b)  (* only one ramp of FS is changed *)
     then begin
      da := 0;
      if ((c + dc) > max_input[V]) or ((c + dc) < b + minwidth)
                               or right_shouldered
        then dc := 0
      else if (c + dc) > (FS_r^.c - (FS_r^.c - FS_r^.b) / 2)
        then dc := FS_r^.c - (FS_r^.c - FS_r^.b) / 2 - c;
     end (* if X > FS.b then *)
     else begin
      dc := 0;
      if ((a + da) < min_input[V]) or ((a + da) > b - minwidth)
                               or left_shouldered
        then da := 0
      else if (a + da) < (FS_l^.b - (FS_l^.b - FS_l^.a) / 2)
        then da := FS_l^.b - (FS_l^.b - FS_l^.a) / 2 - a;
     end; (* if X > b ... else *)

   if ((a + da + minwidth) > (b + db)) or ((c + dc - minwidth) < (b + db))
     then db := 0;

   (* now change FS *)
   a := a + da;
   b := b + db;
   c := c + dc;
   (* and repair the partitioning *)
   if FS_l <> Nil
     then begin
      FS_l^.c := b;
      FS_l^.b := a;
      if Antecedents[V,R] > 2
        then AntecBase[V].FSets[Antecedents[V,R]-2].c := a;
     end;
   if FS_r <> Nil
     then begin
      FS_r^.a := b;
      FS_r^.b := c;
      if Antecedents[V,R] < AntecBase[V].count - 1
        then AntecBase[V].FSets[Antecedents[V,R]+2].a := c;
     end;
  end; (* with *)

 end; (* TNefclass.Constrained_Check_and_Adapt *)

(*-------------------------------------------------------------------------*)
(* This procedure makes sure that by changing a fuzzy set it does not pass *)
(* one of the neighboring fuzzy sets.                                      *)
(* The parameters of a fuzzy set are only changed, if the changes do       *)
(* not conflict with the constraints: a < b < c, and b - a > minwidth, and *)
(* c - b > minwidth. The parameters must also stay within the domain.      *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.NotPass_Check_and_Adapt
                               (R: integer; da, db, dc: real;
                                var min_input, max_input: UnitVector;
                                asym: boolean);
 var
  FS_l, FS, FS_r : ^TriangleFSet;
  V : integer;
  X : real;

 begin

  V := Min_Ant[R];

  (* Select fuzzy set FS and its left and right neighbors, if they exist *)
  if Antecedents[V,R] > 1
    then FS_l := @(AntecBase[V].FSets[Antecedents[V,R]-1])
    else FS_l := Nil;
  FS   := @(AntecBase[V].FSets[Antecedents[V,R]]);
  if Antecedents[V,R] < AntecBase[V].count
    then FS_r := @(AntecBase[V].FSets[Antecedents[V,R]+1])
    else FS_r := Nil;

  X := InLayer[V];

  with FS^ do begin
   if ((a + da) < min_input[V]) or ((a + da) > b - minwidth) or
      left_shouldered or (asym and (x >= b))
     then da := 0;
   if ((c + dc) > max_input[V]) or ((c + dc) < b + minwidth) or
      right_shouldered or (asym and (x <= b))
     then dc := 0;
   if ((a + da + minwidth) > (b + db)) or ((c + dc - minwidth) < (b + db))
     then db := 0;
   (* check neighbors *)
   if FS_l <> nil
     then begin
      if a + da < FS_l^.a then da := FS_l^.a - a;
      if b + db < FS_l^.b then db := FS_l^.b - b;
      if c + dc < FS_l^.c then dc := FS_l^.c - c;
     end;
   if FS_r <> nil
     then begin
      if a + da > FS_r^.a then da := FS_r^.a - a;
      if b + db > FS_r^.b then db := FS_r^.b - b;
      if c + dc > FS_r^.c then dc := FS_r^.c - c;
     end;
   (* now change FS *)
   a := a + da;
   b := b + db;
   c := c + dc;
  end; (* with *)
 end; (* TNefclass.NotPass_Check_and_Adapt *)


(*-------------------------------------------------------------------------*)
(* TNefclass.CheckOutput checks whether a pattern is classified correctly. *)
(* The result is only true, if only the correct output unit has the maximal*)
(* value compared to all other output units. The result will be false, if  *)
(* more than one output unit has the maximal value, or if an incorrect unit*)
(* is maximal. The pattern must be propagated first!                       *)
(*-------------------------------------------------------------------------*)

function TNefclass.CheckOutput (APattern: PPattern; var Class: Integer;
                                var Hit: real; var ambiguous: boolean)
                               : Boolean;
 var
  maximum: real;
  i      : integer;
  correct: boolean;

 begin
  maximum := -1; Class := 0; correct := True;
  for i := 1 to L3 do
   begin
    if OutLayer[i] > maximum
      then begin
       maximum := OutLayer[i];
       Class := i;
      end; (* then *)
   end; (* for i *)
  Hit := maximum;

  ambiguous := false;
  for i := 1 to L3 do
   ambiguous := ambiguous or ((Hit = OutLayer[i]) and (i <> Class));

  CheckOutput := not ambiguous and (Round (APattern^.OutPattern[Class]) = 1)
                 and (Hit > 0);
 end;

(*-------------------------------------------------------------------------*)
(* This procedure does one run through the learning task and creates a rule*)
(* base. For each rule a hidden unit is entered into the network.          *)
(* Rule creation is done by finding for each input variable the fuzzy set  *)
(* giving the highest membership degree for the current input value.       *)
(* If this combination does not exist as an antecedent, a rule unit is     *)
(* created and connected to the output unit that represents the class of   *)
(* current pattern. The fuzzy weights are set accordingly and the          *)
(* conclusion weight is set to 1.  The variable parameters created and     *)
(* wanted show the number of created rules and the number of rules the     *)
(* procedure wanted to create, but was forbidden to do so, because of the  *)
(* restriction given by the maximum number of rules. This is determined by *)
(* the parameter restricted. If it is true, then only TNefclass.max_rules  *)
(* rules will be created. If it is false, then RuleLimit rules will be     *)
(* created. In this case the rule base must be evaluated later, to trim    *)
(* down the number of rule to TNefclass.max_rules.                         *)
(*-------------------------------------------------------------------------*)
procedure TNefclass.CreateRuleBase (var created, wanted: integer; T: PTask;
                                    restricted: boolean);

 (* Local procedure to process a pattern from the collection *)
 procedure Cycle (APattern: PPattern); far;
  var
   i,j,k   : integer;
   Maximum : real;
   MSD     : real;
   MaxMSD  : array[1..maxunits] of real;
   Hit     : real;
   MaxFS   : ByteVector;
   found   : boolean;
   correct : boolean;
   ambiguous: boolean;
   Class   : integer;
   maxrules: integer;

  begin
     if restricted
       then maxrules := max_rules
       else maxrules := RuleLimit;
     (* first check, if pattern is correctly classified *)
     if L2 > 0
       then  begin
        Propagate (APattern, T^.min_input);
        correct := CheckOutput (APattern, Class, Hit, ambiguous);
        (* check if a new rule must be created even if result is correct *)
        if restricted
          then correct := correct and (Hit > 0.0)  (* for simple learning *)
          else correct := correct and (Hit > 0.5); (* for best learning *)
       end
       else begin
        correct := false;
        Hit     := 0;
       end;

     if not correct
       then begin (* A new rule would be nice *)
        for j := 1 to L1 do begin  (* for all inputs find max. ms. degree *)
         Maximum := -1;
         for k := 1 to AntecBase[j].count do begin (* check all fuzzy sets *)
          MSD := Membership(AntecBase[j].Fsets[k],APattern^.InPattern[j]);
          if MSD > Maximum
            then begin
             Maximum  := MSD;
             MaxMSD[j] := MSD;
             MaxFS[j] := k;
            end; (* then *)
         end; (* for k *)
        end; (* for j *)

        (* check if the combination in MaxFS is already antecedent *)
        found := False;
        k := 0;
        while (k < L2) and not found do begin
         inc(k);
         j := 1;
         repeat
          found := (Antecedents[j,k] = MaxFS[j]) or (Antecedents[j,k] = 0);
          inc(j);
         until not found or (j > L1);
        end; (* while *)

        if not found
          then begin (* then we want to create a rule *)
           if L2 < maxrules
             then begin (* then we are allowed to create a rule *)
              inc (L2);
              for j := 1 to L1 do
               Antecedents [j,L2] := MaxFS[j];
              j := 1; k := 0;
              while (j <= L3) and (k = 0) do begin
               if Round(APattern^.OutPattern[j]) = 1 then k := j;
               inc (j);
              end; (* while *)
              ConclusionWeights[L2].node := k;
              ConclusionWeights[L2].weight := 1;
              inc (created);
              RCIndex[created].Pattern := APattern^.Ident;
              RCIndex[created].Rule := L2;
             end (* if L2 ... then *)
             else begin
              inc (wanted); (* we were not allowed to create a rule *)
             end;
          end; (* if not found *)
       end (* if not correct *)
  end; (* Cycle *)

 (* body of CreateRulebase: For each pattern call Cycle *)
 begin
  T^.Patterns^.ForEach(@Cycle);
  if restricted and (L2 > max_rules)
    then L2 := max_rules;
  LearnRules := False;
 end; (* CreateRuleBase *)

(*-------------------------------------------------------------------------*)
procedure TNefclass.ReconnectRules (T: PTask; var EFile: Text);

 var
  RulePerformance : PVectorCollection;
  Vector          : PUnitVector;
  i, j            : integer;
  maximum         : real;
  TheMax          : integer;
  Rule            : integer;

 procedure Cycle (APattern: PPattern); far;
  var
   Class           : integer;
   i,j             : integer;
   Vector          : PUnitVector;
  begin

   Propagate (APattern, T^.min_input);
    for i := 1 to L2 do
     begin
      Vector := PUnitVector(RulePerformance^.At(i-1));
      for j := 1 to L3 do
       Vector^[j] := Vector^[j] + APattern^.OutPattern[j] * RuleLayer[i];
     end;

  end; (* Cycle *)

 begin
  New (RulePerformance, Init (L2,1));
  for i := 1 to L2 do
   begin
    new (Vector);
    for j := 1 to maxunits do
      Vector^[j] := 0;
    RulePerformance^.Insert (Vector);
   end;

  T^.Patterns^.ForEach(@Cycle); (* Check rules with patterns *)

  writeln (EFile, '% Accumulated rule activations per class');
  writeln (EFile);
  for i := 1 to L2 do
   begin
    write (EFile, i: 2, ') ');
    Vector := PUnitVector(RulePerformance^.At(i-1));
    for j := 1 to L3 do
     write (EFile, Vector^[j]:10:4);
    writeln (EFile);
   end;
  writeln (EFile);

  writeln (EFile, '% Change of rule conclusions');
  writeln (EFile);

  for i := 1 to L2 do (* find max. hit ratio for each rule *)
   begin
    maximum := 0; TheMax := 0;
    Vector := PUnitVector(RulePerformance^.At(i-1));
    for j := 1 to L3 do
     if (Vector^[j] > maximum)
       then begin
        maximum := Vector^[j];
        TheMax  := j;
       end;

    if maximum > 0
     then begin
       if ConclusionWeights[i].node <> TheMax
         then begin
          writeln (EFile, 'Conclusion of Rule ', i:2, ' was changed from ',
                   ConclusionWeights[i].node:2, ' to ', TheMax:2, '.');
          ConclusionWeights[i].node := TheMax;
         end
     end
   end; (* for i *)

  Writeln (EFile);
  Dispose(RulePerformance, Done);

 end; (* TNefclass.ReconnectRules *)

(*-------------------------------------------------------------------------*)
procedure TNefclass.EvaluateRules (T: PTask; var EFile: Text;
                                   RLType: RuleLearningType);

 type
  RuleFlags = array[1..RuleLimit] of boolean;

 var
  RulePerformance : ^RuleVector; (* Put this on the heap, not on the stack*)
  RuleSelected    : ^RuleFlags;
  NewAntecedents  : ^WeightMatrix;
  NewConclusions  : ^WeightVector;
  Rule            : integer;
  i, j            : integer;
  chosen          : integer;

 procedure Cycle (APattern: PPattern); far;
  var
   Class           : integer;
   i               : integer;
   Hit             : real;
   Correct         : boolean;
   ambiguous       : boolean;
  begin

   Propagate (APattern, T^.min_input);
   Correct := CheckOutput (APattern, Class, Hit, ambiguous);
    for i := 1 to L2 do
     if correct and (ConclusionWeights[i].node = class)
       then RulePerformance^[i] :=
               RulePerformance^[i]
               + RuleLayer[i]
       else RulePerformance^[i] :=
               RulePerformance^[i]
               - RuleLayer[i];

  end; (* Cycle *)

 procedure FindBest;
  var
   j               : integer;
   maximum         : real;
   TheMax          : integer;
   ready           : boolean;
  begin
   ready := false; chosen := 0;
   while not ready do   (* find best rules *)
    begin
     maximum := -T^.Count; TheMax := 0;
     for j := 1 to L2 do
      if (RulePerformance^[j] > maximum) and not RuleSelected^[j]
        then begin
         maximum := RulePerformance^[j];
         TheMax  := j;
        end;
     if maximum > -T^.Count (* ACHTUNG! *)
      then begin
        RuleSelected^[TheMax] := True;
        inc (chosen);
      end
      else ready := true;
     ready := ready or (chosen >= max_rules);
    end; (* while not ready *)
  end; (* FindBest *)

 procedure FindBestPerClass;
  var
   i, j            : integer;
   class           : integer;
   maximum         : real;
   TheMax          : integer;
   ready           : boolean;
  begin
   class := 1; ready := false; chosen := 0; i := 0;
   while not ready do   (* find best rules per class *)
    begin
     maximum := -T^.Count; TheMax := 0;
     for j := 1 to L2 do
      if (ConclusionWeights[j].node = class) and
         (RulePerformance^[j] > maximum) and not RuleSelected^[j]
        then begin
         maximum := RulePerformance^[j];
         TheMax  := j;
        end;
     if maximum > -T^.Count (* ACHTUNG! *)
      then begin
        RuleSelected^[TheMax] := True;
        inc (chosen); inc (i);
        if (i div class) >= (max_rules div L3)
          then inc(class);
      end
      else begin
       i := (max_rules div L3) * class;
       inc(class);
      end;
     ready := (class > L3);
    end; (* while not ready *)
  end; (* FindBestPerClass *)

 begin
  New (RulePerformance);
  New (RuleSelected);
  New (NewAntecedents);
  New (NewConclusions);
  for i := 1 to RuleLimit do
   begin
    RulePerformance^[i] := 0;
    RuleSelected^[i] := False;
   end;

  T^.Patterns^.ForEach(@Cycle); (* Check rules with patterns *)

  writeln (EFile, '% Overall rule performance');
  writeln (EFile);
  for i := 1 to L2 do
   writeln (EFile, i: 2, ') ', RulePerformance^[i]:10:4);
  writeln (EFile);

  If (RLType = BestPerClass) and (max_rules div L3 >= 1)
    then FindBestPerClass
    else FindBest;

  Rule := 1;
  for i := 1 to L2 do
   begin
    if RuleSelected^[i]
      then begin
       for j := 1 to L1 do
         NewAntecedents^[j,Rule] := Antecedents[j,i];
       NewConclusions^[Rule] := ConclusionWeights[i];
       Inc (Rule);
      end;
   end;
  L2 := chosen;
  for i := 1 to L2 do
   begin
    for j := 1 to L1 do
     Antecedents[j,i] := NewAntecedents^[j,i];
    ConclusionWeights[i] := NewConclusions^[i];
   end;
  Dispose(RulePerformance);
  Dispose(RuleSelected);
  Dispose(NewAntecedents);
  Dispose(NewConclusions);
 end; (* TNefclass.EvaluateRules *)

(*-------------------------------------------------------------------------*)
(* Fuzzy Error Backpropagation: This function returns the error of the     *)
(* current pattern, checks if the pattern is misclassified and carries out *)
(* one adaption step.                                                      *)
(*-------------------------------------------------------------------------*)
function TNefclass.FEBP (APattern: PPattern;
                         var min_input, max_input: UnitVector;
                         LP: PRLearnParam): real;

 var

  PE       : real;                        (* pattern error *)
  i, j     : integer;                     (* loop variables *)
  Class    : integer;                     (* index of max. output unit *)
  Maximum  : real;                        (* max. output value *)
  Delta_out: array [1..RuleLimit] of real; (* delta values of output units *)
  Delta_hid: array [1..RuleLimit] of real; (* delta values of rule units *)
  delta_a  : real;                        (* delta values of the parameters *)
  delta_b  : real;                        (* of the fuzzy sets *)
  delta_c  : real;
  ambiguous: boolean;

 begin

  PE := 0;

  with APattern^, LP^ do begin

   (* determine pattern error and delta values for output units *)
   for i := 1 to L3 do
    begin
     Delta_out[i] := OutPattern[i] - OutLayer[i];
     PE  := PE + sqr(Delta_out[i]);
    end;

   (* determine delta values for hidden units and adapt conclusion weights *)
   for i := 1 to L2 do begin
    if RuleLayer[i] > 0
      then begin (* learning occures only, if the rule is involved *)
       with ConclusionWeights[i] do begin
        Delta_hid[i] := Delta_out[node] * weight *
                        RuleLayer[i] * (1 - RuleLayer[i]);
        if w_learn
          then begin
           weight := weight + s * Delta_out[node] * RuleLayer[i];
           if not free_w_learn
             then if weight > 1
                    then weight := 1
                  else if weight < 0
                    then weight := 0;
          end; (* if w_learning then *)
       end; (* with ConclusionWeights[i] *)

       (* adapt the antecedent of current rule that caused the minimum *)
       with AntecBase[Min_Ant[i]].FSets[Antecedents[Min_Ant[i],i]] do begin
           delta_b :=  sb * Delta_hid[i] * (c - a)
                       * sgn(InLayer[Min_Ant[i]] - b);
           delta_a := -sa * Delta_hid[i] * (b - a) + delta_b;
           delta_c :=  sc * Delta_hid[i] * (c - b) + delta_b;
           if not intersect05
             then begin
              if nopassing
               then NotPass_Check_and_Adapt (
                     i, delta_a, delta_b, delta_c, min_input, max_input,
                     asymmetric)
               else Free_Check_and_Adapt (
                     AntecBase[Min_Ant[i]].FSets[Antecedents[Min_Ant[i],i]],
                     Inlayer[Min_Ant[i]], delta_a, delta_b, delta_c,
                     min_input[Min_Ant[i]], max_input[Min_Ant[i]],
                     asymmetric);
             end
             else Constrained_Check_and_Adapt (
                   i, delta_a, delta_b, delta_c, min_input, max_input);
       end; (* with AntecBase... *)
      end; (* if RuleLayer[i] > 0 then *)
   end; (* for i *)
  end; (* with APattern^, LP^ *)

  (* return the pattern error *)

  FEBP := PE;
 end; (* TNefclass.FEBP *)

(*-------------------------------------------------------------------------*)

procedure TNefclass.Testing (Task: PTask);
 var
  Error    : real;
  MisClass : integer;

 procedure Cycle (APattern: PPattern); far;
  var
   maximum  : real;
   Class    : integer;
   i        : integer;
   perror   : real;
   Hit      : real;
   ambiguous: boolean;
   correct  : boolean;

  begin
   Propagate (APattern, Task^.min_input);
   correct := CheckOutput (APattern, Class, Hit, ambiguous);

   if Task^.ClassInfo
     then begin
      perror := 0;
      for i := 1 to L3 do
       perror := perror + sqr(OutLayer[i] - APattern^.OutPattern[i]);
      Error := Error + perror;
     end;

   if (not Task^.ClassInfo and ((Hit = 0) or ambiguous)) or
      (Task^.ClassInfo and not correct)
     then inc (MisClass)
  end;

 begin
  Error := 0;
  MisClass := 0;
  If L2 > 0
   then begin
    Task^.Patterns^.ForEach(@Cycle);
    Task^.Tested := True;
    Task^.Error := Error;
    Task^.MisClass := MisClass;
   end;
 end; (* TTask.Testing *)


(*-------------------------------------------------------------------------*)

procedure TNefclass.DisplayFuzzyWeights;
 var
   current: integer; (* index of current variable *)
   grdriver, grmode  : integer;
   max_x, max_y, sep_y, x, y, x1, x2, y1, y2, dx, dy, wx, wy : word;
   l : real; (* left boundary of domain (minimum) *)
   s : real; (* scaling factor for width of fuzzy sets *)
   fx, fy : integer;
   i, j, k, kk, n   : integer;
   Atext: string[80];
   VarText: string[10];
   thecolor: word;
   TheKey : char;
   exists, HideFS : Boolean;
   RedrawScreen   : Boolean;

 begin
  (* switch to graphics mode *)
  grdriver := detect;
  InitGraph(grdriver, grmode, PathToBGI);
  ClearViewPort;

  (* draw the screen layout *)
  max_x := GetMaxX; max_y := GetMaxY; sep_y := 3 * max_y div 4;
  x1 := 20; x2 := 620; y1 := 50; y2 := 300;
  dx := 320; dy := 150; wx := 600; wy := 250;
  SetBKColor (LightGray);
  ClearViewPort;
  SetColor (blue);
  SetLineStyle(SolidLn,0,ThickWidth);
  Rectangle (0,0,max_x, max_y);
  Line (0, y1-15, max_x, y1-15);
  SetTextJustify (CenterText, BottomText);
  SetTextStyle (TriplexFont,HorizDir,2);
  MoveTo (max_x div 2 , (y1-25));
  SetColor(red);
  OutText ('NEFCLASS-PC');
  SetTextJustify (RightText, BottomText);
  SetTextStyle (SmallFont,HorizDir,5);
  MoveTo (max_x - 5 , (y1-25));
  OutText ('(c) TU Braunschweig, 1994');
  SetTextJustify (LeftText, BottomText);
  SetTextStyle (SmallFont,HorizDir,5);
  MoveTo (5 , (y1-25));
  OutText ('Version ' + VersionString + ', ' + DateString);
  SetTextStyle (DefaultFont,HorizDir,1);
  SetColor(Blue);
  y := sep_y;
  Line (0, y, max_x, y);
  y := max_y - 40;
  Line (0, y, max_x, y);
  SetTextJustify (LeftText, BottomText);
  MoveTo (2, y + 12);
  OutText (' Display Mode: Use PgUp/PgDown to scroll through'+
           ' the input variables.');
  MoveTo (2, y + 24);
  OutText ('               H hides/reveals the fuzzy sets'+
           ' not used within the rule base.');
  MoveTo (2, y + 36);
  OutText ('               ESC exits display mode.');

  str (0: 6, VarText);
  AText := 'Epoch: ' + VarText;
  str (0.0:10:6, VarText);
  AText := AText + ', Error: ' + VarText;
  str (0.0:10:6, VarText);
  AText := AText + ', mean error: ' + VarText;
  OutTextXY (20, sep_y + 30, AText);

  str (0, VarText);
  AText := VarText + ' patterns out of ';
  str (0, VarText);
  AText := AText + VarText + ' were misclassified';
  OutTextXY (20, sep_y + 60, AText);
  current := 1; HideFS := False;
  RedrawScreen := False;

  repeat
   SetColor (blue);
   SetLineStyle (SolidLn, 0 , ThickWidth);
   Line (x1, y1, x1, y2);
   Line (x1, y2, x2, y2);
   If HideFS
     then AText := AntecBase[current].VName +
                   '  (Unused fuzzy sets will be hidden)'
     else AText := AntecBase[current].VName +
                   '  (Dotted Lines: unused fuzzy sets)';
   OutTextXY (x1, y2+30, AText);
   str (InVarRange[current].min:0:2, VarText);
   OutTextXY (x1, y2+15, VarText);
   SetTextJustify (RightText, BottomText);
   str (InVarRange[current].max:0:2, VarText);
   OutTextXY (x2, y2+15, VarText);
   SetTextJustify (LeftText, BottomText);
   with AntecBase[current] do begin
    l := InVarRange[current].min;
    s := wx / (InVarRange[current].max - l);
    for j := 1 to count do
     begin
      exists := False;
      for k := 1 to L2 do
       exists := exists or (Antecedents[current,k] = j);

      with FSets[j] do begin
       thecolor := (white - j) mod 16;
       if thecolor = GetBKColor
         then thecolor := (thecolor - 1) mod 16;
       SetColor (thecolor);
       if exists or not HideFS
         then begin
          if exists
            then SetLineStyle (SolidLn, 0 , ThickWidth)
            else SetLineStyle (DottedLn, 0 , ThickWidth);
          if left_shouldered
            then y := y1  (* dms = 1 (Degree of MemberShip) *)
            else y := y2; (* dms = 0 *)
          MoveTo (x1 + round ((a - l) * s), y);
          LineTo (x1 + round ((b - l) * s), y1); (* dms = 1 *)
          if right_shouldered
            then y := y1  (* dms = 1 *)
            else y := y2; (* dms = 0 *)
          LineTo (x1 + round ((c - l) * s), y);
         end; (* then *)
      end; (* with FSets *)
     end; (* for j *)
   end; (* with AntecBase *)


   TheKey := UpCase(ReadKey);
   Case TheKey of
    #0 : begin
          TheKey := ReadKey;
          Case TheKey of
           #73: if current < L1 then inc (current); (* PgUp   *)
           #81: if current > 1  then dec (current);   (* PgDown *)
          end; (* case *)
          RedrawScreen := True;
         end; (* #0 *)
    'H': begin
          HideFS := Not HideFS;
          RedrawScreen := True;
         end; (* H *)
   end; (* Case *)

   if RedrawScreen
     then begin
      SetViewPort (x1,y1-5,x2,y2+30,ClipOn);
      ClearViewPort;
      SetViewPort (0,0,max_x,max_y,ClipOn);
      RedrawScreen := False;
     end;
  until TheKey = #27;

  (* switch back to text mode, and redraw desktop *)
  CloseGraph;
  Application^.SetScreenMode(smCO80);
  Application^.ReDraw;
 end; (* TNefclass.DisplayFuzzyWeights *)


end.