(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NFCL_TV.PAS                                                   *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 08/16/95                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

program NEFCLASS;

uses Memory, NFCL_Con, Drivers, Objects, Views, Menus, App, Dialogs,
  Editors, StdDlg, Validate, MyValid, MsgBox, TextView, HelpFile,
  NFCL_Dlg, NFCL_Typ, NFCL_Tsk, NFCL_Obj, NFCL_Run, NFCLHELP;

type

 NFCL_App = object (TApplication)
     ClipboardWindow     : PEditWindow;
     LearnParamDialog    : PLearnParamDialog;
     LearnParamBuffer    : PRLearnParamBuf;
     LearnParam          : PRLearnParam;
     NetParamDialog      : PNetParamDialog;
     NetParamBuffer      : PRNetParam;
     NewNet              : Boolean;
     LTask               : PTask;
     Net                 : PNefclass;
     constructor Init;
     destructor Done; virtual;
     function GetPalette: PPalette; virtual;
     procedure DoAboutBox;
     procedure DoLicenceBox;
     procedure Learning;
     procedure Testing;
     procedure Statistics;
     procedure OpenLearnParamDialog;
     procedure SetLearnParams;
     procedure GetLearnParams;
     function  OpenNetParamDialog (HelpContext:integer): Boolean;
     procedure SetNetParams;
     procedure NewNetwork;
     procedure LoadNetwork;
     procedure SaveNetwork;
     procedure WriteNetRules;
     procedure ShowNetParams;
     procedure LoadPattern;
     procedure DisplayPattern;
     procedure GetEvent(var Event: TEvent); virtual;
     procedure HandleEvent(var Event: TEvent); virtual;
     procedure InitMenuBar; virtual;
     procedure InitStatusLine; virtual;
     procedure LoadDesktop;
     procedure NewWindow;
     procedure OpenWindow;
     procedure SaveDesktop;
 end;

var
 NFC           : NFCL_App;
 ResFile       : TResourceFile;
 NetFile       : Text;
 PatternFile   : Text;
 RuleFile      : Text;

(*-------------------------------------------------------------------------*)
(*                        Methods of NFCL_App                              *)
(*-------------------------------------------------------------------------*)

procedure NFCLStreamError(var S: TStream); far;
var
  ErrorMessage: String;
begin
  case S.Status of
    stError: ErrorMessage := 'Stream access error';
    stInitError: ErrorMessage := 'Cannot initialize stream';
    stReadError: ErrorMessage := 'Read beyond end of stream';
    stWriteError: ErrorMessage := 'Cannot expand stream';
    stGetError: ErrorMessage := 'Unregistered type read from stream';
    stPutError: ErrorMessage := 'Unregistered type written to stream';
    end;
  DoneVideo;
  PrintStr('Error: ' + ErrorMessage);
  Halt(Abs(S.Status));
end;

(*-------------------------------------------------------------------------*)

constructor NFCL_App.Init;
var
  R: TRect;
begin
  MaxHeapSize := 8192;
  EditorDialog := StdEditorDialog;
  StreamError := @NFCLStreamError;
  RegisterMenus;
  RegisterObjects;
  RegisterViews;
  RegisterApp;
  RegisterEditors;
  RegisterHelpFile;
  RegisterDialogs;
  RegisterValidate;
  RegisterMyValidate;
  RegisterMyDialogs;
  ResFile.Init(New(PBufStream, Init('NEFCLASS.TVR', stOpenRead, 1024)));
  inherited Init;
  Desktop^.GetExtent(R);
  ClipboardWindow := New(PEditWindow, Init(R, '', wnNoNumber));
  if ValidView(ClipboardWindow) <> nil then
  begin
    ClipboardWindow^.Hide;
    InsertWindow(ClipboardWindow);
    Clipboard := ClipboardWindow^.Editor;
    Clipboard^.CanUndo := False;
  end;
  LearnParam := nil;
  NewNet     := True;
  LTask      := nil;
  Net        := nil;
  SetLearnParams;
  GetLearnParams;
  DoAboutBox;
  DoLicenceBox;
end;

(*-------------------------------------------------------------------------*)

destructor NFCL_App.Done;
begin
  if LearnParamBuffer <> nil then Dispose (LearnParamBuffer);
  if NetParamBuffer   <> nil then Dispose (NetParamBuffer);
  if LearnParam       <> nil then Dispose (LearnParam);
  if LTask            <> nil then LTask^.Done;
  if Net              <> nil then Net^.Done;
  ResFile.Done;
  inherited Done;
end;

(*-------------------------------------------------------------------------*)

function NFCL_App.GetPalette: PPalette;
const
  CNewColor = CAppColor + CHelpColor;
  CNewBlackWhite = CAppBlackWhite + CHelpBlackWhite;
  CNewMonochrome = CAppMonochrome + CHelpMonochrome;
  P: array[apColor..apMonochrome] of string[Length(CNewColor)] =
    (CNewColor, CNewBlackWhite, CNewMonochrome);
begin
  GetPalette := @P[AppPalette];
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.OpenLearnParamDialog;
 begin
  LearnParamDialog := PLearnParamDialog(ResFile.Get('PARAMETER'));
  LearnParamDialog^.HelpCTX := hcParamDialog;
  if LearnParamBuffer = nil then SetLearnParams;
  if ExecuteDialog (LearnParamDialog, LearnParamBuffer) <> cmCancel
    then GetLearnParams;
 end; (* NFCL_App.OpenLearnParamDialog *)

(*-------------------------------------------------------------------------*)
(**)
procedure NFCL_App.GetLearnParams;
 var
  Code: integer;
 begin
  if LearnParam = nil then new (LearnParam);
  with LearnParamBuffer^, LearnParam^ do begin
    val (Epochs, EndEpoch, Code);
    val (StopErr, StopError, Code);
    val (NofErrors, MisClass_OK, Code);
    intersect05 := Intersect = 0;
    w_learn    := FixedConc = 1;
    free_w_learn := UnitConc = 1;
    RuleLearning := RuleLearningType(RuleCreation);
    single_step := false;
    val (Update, UpdateRate, Code);
    TextDisplay := True;
    StopCheck := StopCriterion <> 0;
    val (ErrorEpochs, ErrEpochs, Code);
    val (sigma_a, sa, Code);
    val (sigma_b, sb, Code);
    val (sigma_c, sc, Code);
    val (sigma,   s, Code);
  end;(* with *)
 end; (* NFCL_App.GetLearnParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SetLearnParams;
 begin
  if LearnParamBuffer = nil then New (LearnParamBuffer);
  with LearnParamBuffer^ do begin
    Epochs   := '500';
    StopErr  := '0.0';
    NofErrors:= '0';
    Intersect:= 1;
    FixedConc:= 0;
    UnitConc := 0;
    RuleCreation := 0;
    StopCriterion := 0;
    ErrorEpochs := '100';
    Update   := '10';
    sigma_a  := '0.001';
    sigma_b  := '0.001';
    sigma_c  := '0.001';
    sigma    := '0.001';
  end; (* with *)
 end; (* NFCL_App.SetLearnParams *)

(*-------------------------------------------------------------------------*)
procedure NFCL_App.Learning;
 var
   Trainer: PTrainer;
 begin
  if (Net <> nil) and (LTask <> nil)
    then begin
     if (Net^.L1 > 0) and (LTask^.count > 0)
       then begin
        Trainer := New(PTrainer, Init (LearnParam));
        Trainer^.HelpCTX := hcLearnDialog;
        if ValidView(Trainer) <> nil
          then begin
           InsertWindow (Trainer);
           Trainer^.Training (Net, LTask, NewNet);
           NewNet := False;
          end;
       end;
    end
    else MessageBox (#3'You must first define a network'#13#3+
                       'and patterns!', nil,
                     mfError or mfOKButton);
 end; (* NFCL_App.Learning *)

(*-------------------------------------------------------------------------*)
procedure NFCL_App.Testing;
 var
   Tester: PTester;
 begin
  if (Net <> nil) and (LTask <> nil)
    then begin
     if (Net^.L1 > 0) and (LTask^.count > 0)
       then begin
        Tester := New(PTester, Init (Net, LTask));
        Tester^.HelpCTX := hcTestDialog;
        if ValidView(Tester) <> nil
          then begin
           InsertWindow (Tester);
          end;
       end
    end
    else MessageBox (#3'You must first define a network'#13#3+
                       'and patterns!', nil,
                     mfError or mfOKButton);
 end; (* NFCL_App.Testing *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.Statistics;
 var
  Stats: PStatistics;
 begin
  if (Net <> nil) and (LTask <> nil)
    then begin
     if (Net^.L1 > 0) and (LTask^.count > 0)
       then begin
        Stats := New(PStatistics, Init (Net, LTask));
        if ValidView(Stats) <> nil
          then InsertWindow (Stats);
       end
    end
    else MessageBox (#3'You must first define a network'#13#3+
                       'and patterns!', nil,
                     mfError or mfOKButton);
 end; (* NFCL_App.Testing *)

(*-------------------------------------------------------------------------*)


function NFCL_App.OpenNetParamDialog (HelpContext:integer): Boolean;
 var
   i          : integer; (* loop variable *)
   Code       : integer; (* code for interpreting string as value *)
   L1, L3     : integer; (* number of units in input and output layer *)
   nofs       : integer; (* number of fuzzy sets per unit *)
   maxrules   : integer; (* maximum number of units in rule layer *)
   OK         : boolean; (* true, if new parameters entered *)
   UseSum     : boolean; (* true, if aggregation by weighted sum *)
   Result     : word;    (* Result of ExecuteDialog (button pressed) *)
 begin
  OK := False;
  NetParamDialog := PNetParamDialog(ResFile.Get('NETPARAM'));
  NetParamDialog^.HelpCTX := HelpContext;
  if NetParamBuffer = nil then SetNetParams;
  Result := ExecuteDialog (NetParamDialog, NetParamBuffer);
  if  Result = cmOK
    then begin
     val (NetParamBuffer^.InLayer, L1, Code);
     val (NetParamBuffer^.RuleLayer, maxrules, Code);
     val (NetParamBuffer^.OutLayer, L3, Code);
     val (NetParamBuffer^.Partition, nofs, Code);
     UseSum := NetParamBuffer^.UseSum = 0;
     if Net <> nil then Net^.Done;
     New (Net, Init(L1, maxrules, L3, nofs, UseSum));
     OK := True;
    end;
  (* Return True, if new net was defined, or if weights must be displayed *)
  OpenNetParamDialog := OK or (Result = cmYes);
 end; (* NFCL_App.OpenNetParamDialog *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SetNetParams;
 begin
  if NetParamBuffer = nil then New (NetParamBuffer);
  with NetParamBuffer^ do begin
   if Net <> nil
     then begin
      str (Net^.L1, InLayer);
      if NewNet
        then str (Net^.max_rules, RuleLayer)
        else str (Net^.L2, RuleLayer);
      str (Net^.L3, OutLayer);
      str (Net^.Antecbase[1].count, Partition);
      if Net^.UseSum
        then UseSum := 0
        else UseSum := 1;
     end
     else begin
      InLayer   := '4';
      RuleLayer := '7';
      OutLayer  := '3';
      Partition := '3';
      UseSum    := 0;
     end;
  end;
 end; (* NFCL_App.SetNetParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.NewNetwork;
 begin
  if Net <> nil
    then begin
     if MessageBox (#3'There is already a network defined!'#13#3+
                        'This network will be deleted.'#13#3+
                        'OK to proceed?', nil,
                        mfInformation or mfOKButton or mfCancelButton)
        <> cmCancel
       then begin
         Net^.Done;
         Net := nil;
         SetNetParams;
         DisableCommands([cmYes]);
         NewNet := OpenNetParamDialog (hcNetNew);
         EnableCommands([cmYes]);
       end;
    end
    else begin
     DisableCommands([cmYes]);
     NewNet := OpenNetParamDialog (hcNetNew);
     EnableCommands([cmYes]);
    end;
 end; (* NFCL_App.NewNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.LoadNetwork;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
  OK        : Boolean;
const
  FDOptions: Word = fdOKButton or fdOpenButton or fdHelpButton;
begin
  if Net <> nil
    then OK := (MessageBox (#3'There is already a network defined!'#13#3+
                        'This network will be deleted.'#13#3+
                        'OK to proceed?', nil,
                        mfInformation or mfOKButton or mfCancelButton)
                <> cmCancel)
     else OK := True;
  if OK
    then begin
     if Net <> nil then Net^.Done;
     TheFile := '*.net';
     New(FileDialog, Init(TheFile, 'Open network file', '~F~ile name',
       FDOptions, 1));
     FileDialog^.HelpCTX := hcNetLoad;
     if ExecuteDialog(FileDialog, @TheFile) <> cmCancel
       then begin
        if Filename (NetFile, reading, TheFile)
          then begin
           New (Net, Init(0,0,0,0,true));
           if Net^.ReadNet (NetFile)
                 then begin
                   MessageBox ('Network file was read successfully', nil,
                                mfInformation or mfOKButton);
                   NewNet := False;
                   SetNetParams;
                 end
                 else begin
                  Net^.Done;
                  Net := nil;
                  MessageBox ('Could not read Network file', nil,
                              mfError or mfOKButton);
                 end;
           end; (* if Net^.ReadNet then *)
       end; (* if Filename then *)
    end; (* if ExecuteDialog then *)
end; (* NFCL_App.LoadNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SaveNetwork;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
const
  FDOptions: Word = fdOKButton or fdHelpButton;
begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                     mfError or mfOKButton)
    else begin
     TheFile := '*.net';
     New(FileDialog, Init(TheFile, 'Save network file as', '~F~ile name',
       FDOptions, 1));
     FileDialog^.HelpCTX := hcNetSave;
     if ExecuteDialog(FileDialog, @TheFile) <> cmCancel then
     begin
      if Filename (NetFile, writing, TheFile)
        then begin
         Net^.WriteNet (NetFile, 'Testing');
         MessageBox ('Network file was written successfully', nil,
                     mfInformation or mfOKButton)
        end
        else MessageBox ('Could not write Network file', nil,
                         mfError or mfOKButton);
    end; (* else *)
  end;
end; (* NFCL_App.SaveNetwork *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.WriteNetRules;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
const
  FDOptions: Word = fdOKButton or fdHelpButton;
begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                     mfError or mfOKButton)
    else begin
     if Net^.L2 <= 0
       then MessageBox (#3'The network has no rules yet!', nil,
                     mfError or mfOKButton)
       else begin
        TheFile := '*.rul';
        New(FileDialog, Init(TheFile, 'Save rules in file', '~F~ile name',
          FDOptions, 1));
        FileDialog^.HelpCTX := hcNetRules;
        if ExecuteDialog(FileDialog, @TheFile) <> cmCancel then
        begin
         if Filename (RuleFile, writing, TheFile)
           then begin
            Rewrite (RuleFile);
            Net^.WriteRuleBase (RuleFile);
            Close (RuleFile);
            MessageBox ('Rule base was written successfully!', nil,
                        mfInformation or mfOKButton)
           end
           else MessageBox ('Could not write rule base file!', nil,
                            mfError or mfOKButton);
       end; (* else *)
    end; (* else *)
  end;
end; (* NFCL_App.WriteNetRules *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.ShowNetParams;
 var
  OK : Boolean;
 begin
  if Net = nil
    then MessageBox (#3'There is no network defined!', nil,
                      mfError or mfOKButton)
    else begin
     DisableCommands ([cmOK]);
     OK := OpenNetParamDialog (hcNetDisplay);
     EnableCommands ([cmOK]);
     if OK and (Net^.L2 > 0) then Net^.DisplayFuzzyWeights;
    end;
 end; (* NFCL_App.ShowNetParams *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.LoadPattern;
var
  FileDialog: PFileDialog;
  TheFile   : FNameStr;
  OK        : Boolean;
const
  FDOptions: Word = fdOKButton or fdOpenButton or fdHelpButton;
begin
  OK := True;
  if LTask = nil then new(LTask, Init);
  if LTask^.Count > 0
    then OK := (MessageBox (#3'There are already patterns defined!'#13#3+
                        'These patterns will be deleted.'#13#3+
                        'OK to proceed?', nil,
                        mfInformation or mfOKButton or mfCancelButton)
                <> cmCancel);
  if Net = nil
    then begin
     OK := False;
     MessageBox (#3'You must first define a network!', nil,
                 mfError or mfOKButton);
    end;
  if OK
    then begin
     TheFile := '*.dat';
     New(FileDialog, Init(TheFile, 'Open pattern file', '~F~ile name',
       FDOptions, 1));
     FileDialog^.HelpCTX := hcPatLoad;
     if ExecuteDialog(FileDialog, @TheFile) <> cmCancel
       then begin
        if Filename (PatternFile, reading, TheFile)
          then if LTask^.ReadPattern (PatternFile, Net^.L1, Net^.L3)
                 then begin
                   MessageBox (#3'Pattern file was read successfully', nil,
                                mfInformation or mfOKButton);
                 end
                 else MessageBox (#3'Pattern file does not fit'#13#3+
                                  'to network',
                                   nil, mfError or mfOKButton);
       end;
    end;
end; (* NFCL_App.LoadPattern *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DisplayPattern;
 var
  TermWin     : PTermWin;
  Term        : PText;
  i           : integer;
 begin
  if LTask <> nil
    then begin
     New(TermWin, Init ('Parameters of loaded patterns', 41, 12));
     if ValidView(TermWin) <> nil
       then begin
        InsertWindow (TermWin);
        TermWin^.GetTerm (Term);
        writeln (Term^, 'Number of Patterns      : ', LTask^.Count);
        writeln (Term^, 'Number of input features: ', LTask^.Nin);
        writeln (Term^, 'Number of classes       : ', LTask^.Nout);
        writeln (Term^);
        writeln (Term^, 'Ranges of the input features:');
        for i := 1 to LTask^.Nin do
          writeln (Term^, i:2, ') Min = ', LTask^.min_input[i]:7:2,
                   ', Max = ', LTask^.max_input[i]:7:2);
       end; (* then *)
    end (* then *)
    else
     MessageBox (#3'There are no patterns defined',nil,
                 mfError or mfOkButton);
 end; (* NFCL_App.DisplayPattern *)

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DoAboutBox;
begin
  ExecuteDialog(PDialog(ResFile.Get('ABOUTBOX')), nil);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.DoLicenceBox;
begin
  ExecuteDialog(PDialog(ResFile.Get('LICENCEBOX')), nil);
end;

(*-------------------------------------------------------------------------*)
procedure NFCL_App.GetEvent(var Event: TEvent);
var
  W        : PWindow;
  HFile    : PHelpFile;
  HelpStrm : PDosStream;
const
  HelpInUse: Boolean = False;
begin
  inherited GetEvent(Event);
  case Event.What of
    evCommand:
      if (Event.Command = cmHelp) and not HelpInUse then
      begin
        HelpInUse := True;
        HelpStrm  := New(PDosStream, Init(CalcHelpName, stOpenRead));
        HFile     := New(PHelpFile, Init(HelpStrm));
        if HelpStrm^.Status <> stOk then
        begin
          MessageBox('Could not open help file.', nil, mfError + mfOkButton);
          Dispose(HFile, Done);
        end
        else
        begin
          W := New(PHelpWindow,Init(HFile, GetHelpCtx));
          if ValidView(W) <> nil then
          begin
            ExecView(W);
            Dispose(W, Done);
          end;
          ClearEvent(Event);
        end;
        HelpInUse := False;
      end;
    evMouseDown:
      if Event.Buttons <> 1 then Event.What := evNothing;
  end;
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.HandleEvent(var Event: TEvent);
var
  R: TRect;
begin
  inherited HandleEvent(Event);
  if Event.What = evCommand then
  begin
    case Event.Command of
      cmOptionsLoad:
        begin
          LoadDesktop;
          ClearEvent(Event);
        end;
      cmOptionsSave:
        begin
          SaveDesktop;
          ClearEvent(Event);
        end;
      cmClipShow:
        with ClipboardWindow^ do
        begin
          Select;
          Show;
          ClearEvent(Event);
        end;
      cmNew:
        begin
          NewWindow;
          ClearEvent(Event);
        end;
      cmOpen:
        begin
          OpenWindow;
          ClearEvent(Event);
        end;
      cmNetworkNew     : NewNetwork;
      cmNetworkLoad    : LoadNetwork;
      cmNetworkSave    : SaveNetwork;
      cmNetworkDisplay : ShowNetParams;
      cmNetworkRules   : WriteNetRules;
      cmPatternLoad    : LoadPattern;
      cmPatternDisplay : DisplayPattern;
      cmRunLearn       : Learning;
      cmRunParam       : OpenLearnParamDialog;
      cmRunTest        : Testing;
      cmRunStats       : (*Statistics*);
      cmOptionsVideo:
        begin
          SetScreenMode(ScreenMode xor smFont8x8);
          ClearEvent(Event);
        end;
      cmAbout:
        begin
          DoAboutBox;
          ClearEvent(Event);
        end;
      cmLicence:
        begin
          DoLicenceBox;
          ClearEvent(Event);
        end;
    end;
  end;
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.InitMenuBar;
begin
  MenuBar := PMenuBar(ResFile.Get('MAINMENU'));
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.InitStatusLine;
var
  R: TRect;
begin
  StatusLine := PStatusLine(ResFile.Get('STATUS'));
  GetExtent(R);
  StatusLine^.MoveTo(0, R.B.Y - 1);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.LoadDesktop;
var
  DesktopFile: TBufStream;
  TempDesktop: PDesktop;
  R: TRect;
begin
  DesktopFile.Init('NEFCLASS.DTP', stOpenRead, 1024);
  TempDesktop := PDesktop(DesktopFile.Get);
  DesktopFile.Done;
  if ValidView(TempDesktop) <> nil then
  begin
    Desktop^.Delete(ClipboardWindow);
    Delete(Desktop);
    Dispose(Desktop, Done);
    Desktop := TempDesktop;
    Insert(Desktop);
    GetExtent(R);
    R.Grow(0, -1);
    Desktop^.Locate(R);
    InsertWindow(ClipboardWindow);
  end;
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.NewWindow;
var
  R: TRect;
  TheWindow: PEditWindow;
begin
  R.Assign(0, 0, 60, 20);
  TheWindow := New(PEditWindow, Init(R, '', wnNoNumber));
  InsertWindow(TheWindow);
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.OpenWindow;
var
  R: TRect;
  FileDialog: PFileDialog;
  TheFile: FNameStr;
const
  FDOptions: Word = fdOKButton or fdOpenButton or fdHelpButton;
begin
  TheFile := '*.*';
  New(FileDialog, Init(TheFile, 'Open file', '~F~ile name',
    FDOptions, 1));
  FileDialog^.HelpCTX := hcFOFileOpenDBox;
  if ExecuteDialog(FileDialog, @TheFile) <> cmCancel then
  begin
    R.Assign(0, 0, 75, 20);
    InsertWindow(New(PEditWindow, Init(R, TheFile, wnNoNumber)));
  end;
end;

(*-------------------------------------------------------------------------*)

procedure NFCL_App.SaveDesktop;
var
  DesktopFile: TBufStream;
begin
  Desktop^.Delete(ClipboardWindow);
  DesktopFile.Init('NEFCLASS.DTP', stCreate, 1024);
  DesktopFile.Put(Desktop);
  DesktopFile.Done;
  InsertWindow(ClipboardWindow);
end;

(*-------------------------------------------------------------------------*)

begin
 NFC.Init;
 NFC.Run;
 NFC.Done;
end.